/*
 * main.c -- icebuteo
 *
 * $Revision: 47 $
 * $Date: 2007-05-27 12:12:40 +0200 (Sun, 27 May 2007) $
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "icebuteo.h"
#include "xwindow.h"
#include "vars.h"
#include "config.h"
#include "modules.h"

#define DEBUG 0
#include "debug.h"

int the_time = 0;
int done = 0, respawn = 0;
int toggle_visibility = 0;

time_t start_time = 0;

void on_signal(int sig) {
	switch (sig) {
		case SIGINT:
		case SIGTERM:
			done = 1;
			break;

		case SIGALRM:
			toggle_visibility = 1;
			break;
	
		case SIGHUP:
			/* do not respawn too soon, because colisions in xdm auth */
			if (start_time + 1 < time(NULL)) { 
				done = 1;
				respawn = 1;
			}
			break;
	}
}

int main (int argc, char *argv[], char *envp[]) {
	XWINDOW *w;
	MODULE *mod;
	fd_set rd;
	struct timeval tv;
	int err;
	int s, l;
	char *str;
	long long timeout_duration;
	int old_cwd_fd;

	/* store cwd */
	old_cwd_fd = open(".", O_RDONLY);
	if (old_cwd_fd == -1) {
		perror(PROG_NAME ": Can't open current directory");
	}

	/* local time & locales */
	tzset();
	setlocale(LC_ALL, "");
	setlocale(LC_NUMERIC, "POSIX");		/* numbers in configuration */

	/* signal handlers */
	start_time = time(NULL);
	signal(SIGINT, on_signal);
	signal(SIGTERM, on_signal);
	signal(SIGALRM, on_signal);
	signal(SIGHUP, on_signal);

	/* load configuration */
	if (argc == 2) {
		l = 0;
		str = argv[1];
	} else {
		l = strlen(getenv("HOME")) + strlen(DEFAULT_CONFIG) + 2;
		if (!(str = (char*) malloc(l))) {
			perror(PROG_NAME);
			abort();
		}
		strcpy(str, getenv("HOME"));
		strcat(str, "/");
		strcat(str, DEFAULT_CONFIG);
	}
	if (!config_load(str)) {
		return EXIT_FAILURE;
	}
	if (l)
		free(str);
	
	/* initialisation */
	var_init();
	w = xwindow_init();
	if (!w)
		return EXIT_FAILURE;
	xwindow_load_widgets(w);

	mod = module_init(cfg.module_path);
	if (module_load_config(mod) == -1) {
		fprintf(stderr, "%s: Modules configuration failed.\n", PROG_NAME);
		return EXIT_FAILURE;
	}
	if (module_load_all(mod) == -1) {
		fprintf(stderr, "%s: Module loading failed.\n", PROG_NAME);
		return EXIT_FAILURE;
	}

	timeout_duration = module_handle_timeout(mod, 0, 0, 0);
	memset(&tv, 0, sizeof(struct timeval));
	tv.tv_usec = timeout_duration;

	/* preset autohide-on-start */
	if (cfg.autohide_on_start > 0) {
		alarm(cfg.autohide_on_start);
	}

	/* main loop */
	DBG_MSG(1, "Entering main loop ...");
	while (!done && !xwindow_is_broken(w)) {
		DBG_MSG(5, "before select");
		/* select init */
		FD_ZERO(&rd);
		FD_SET(xwindow_get_fd(w), &rd);
		module_fill_fdset(mod, &rd);
		
		/* visibility update */
		if (toggle_visibility) {
			xwindow_visible(w, -1);
			toggle_visibility = 0;
			if (w->is_visible)
				timeout_duration = module_handle_timeout(mod, timeout_duration, 0, 1);
		}

		DBG_MSG(5, "select ...");
		s = select(FD_SETSIZE, &rd, NULL, NULL, timeout_duration ? &tv : NULL);
		err = errno;
	
		if (s == -1) {
			/* error */
			DBG_MSG(5, "-> error \"%s\"", strerror(err));
			if (err != EINTR) {
				errno = err;
				perror(PROG_NAME ": select");
			}
		} else {
			if (s > 0) {
				/* data */
				DBG_MSG(5, "-> data");

				/* X server connection */
				if (FD_ISSET(xwindow_get_fd(w), &rd)) {
					/* X window event */
					xwindow_process_events(w);
				}

				/* fd callbacks */
				module_handle_fds(mod, &rd);

				/* update timeout */
				timeout_duration = timeout_duration - ((long long)tv.tv_sec * 1e6 + tv.tv_usec);
			} else {
				DBG_MSG(5, "-> timeout");
			}

			/* timeout */
			timeout_duration = module_handle_timeout(mod, timeout_duration, !(w->is_visible), 0);
			memset(&tv, 0, sizeof(struct timeval));
			tv.tv_usec = timeout_duration;

			/* redraw window */
			if (var_changed(1)) {
				xwindow_redraw(w, var_visibility_changed(1));
			}
		}
	}
	DBG_MSG(1, "Main loop done.");
	
	/* cleaning up */
	module_destroy(mod);
	xwindow_destroy(w);
	var_destroy();

	/* May I die ? */
	if (respawn) {
		DBG_MSG(1, "Trying to respawn...");
		if (old_cwd_fd != -1) {
			if (fchdir(old_cwd_fd) == -1) {
				perror(PROG_NAME ": Can't return to origin directory");
				return EXIT_FAILURE;
			}
			DBG_MSG(1, "==========================================");
			DBG_MSG(1, "   Respawning now ...");
			DBG_MSG(1, "==========================================");
			execvp(argv[0], argv);
			perror(PROG_NAME ": execvp failed");
			return EXIT_FAILURE;
		} else {
			fprintf(stderr, PROG_NAME ": Can't respawn. Exiting.\n");
			return EXIT_FAILURE;
		}
	}
	
	return EXIT_SUCCESS;
}

