/*
 * main.c - icebuteo module "Battery-sys"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

#include "battery.h"

/* debug */
#define DEBUG 0
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and one integer parameter.
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module real name */
const char *mod_name = "Battery";

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */
	char *battery;
	int refresh;
	int show_hours;

	/* configuration */
	int vo_state_string;
	int vo_percent;
	int vo_float;
	int vo_time_remaining;
} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))


/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}

/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	int i = 1;
	const int cfg_tr_len = 8;
	P_TRANS cfg_tr[] = {
		/* const. name,			id, var, type */
		{ "battery",			2, NULL, P_STRING },
		{ "refresh",			7, NULL, P_INTEGER },
		{ "show-hours",			8, NULL, P_INTEGER },
		{ "var-float",			5, NULL, P_INTEGER },
		{ "var-percent",		4, NULL, P_INTEGER },
		{ "var-state-string",		3, NULL, P_INTEGER },
		{ "var-time-remaining",		6, NULL, P_INTEGER },
		{ NULL, 0, NULL, 0 }
	};
	static P_FORMAT cfg_format[] = {
		P_CONSTANT, P_EQUAL, P_STRING | P_INTEGER | P_FLOAT, 0
	};

	DBG_MSG(1, "initialize...");

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));
	memset((void *)*mod_data, 0, sizeof(MOD_DATA));
	MD->mod_filedescriptor_value = NULL;
	MD->show_hours = 1;

	/* save offset */
	MD->var_offset = var_offset;

	/* load configuration */
	i = 0;
	cfg_tr[i++].var = &(MD->battery);
	cfg_tr[i++].var = &(MD->refresh);
	cfg_tr[i++].var = &(MD->show_hours);
	cfg_tr[i++].var = &(MD->vo_float);
	cfg_tr[i++].var = &(MD->vo_percent);
	cfg_tr[i++].var = &(MD->vo_state_string);
	cfg_tr[i++].var = &(MD->vo_time_remaining);
	if (!parse_file(arg, NULL, cfg_format, -1, 0, 2, cfg_tr, cfg_tr_len)) {
		return -1;
	}

	/* register timeout - first after 0.1 seconds */
	MD->mod_timeout_value = 100e3;

	/* init vars */
	var_set_as_int(MD->var_offset + 0, 0);
	var_set_as_float(MD->var_offset + MD->vo_float, 0.0);
	var_set_as_string(MD->var_offset + MD->vo_percent, "-- %");
	var_set_as_string(MD->var_offset + MD->vo_state_string, "unknown");
	var_set_as_string(MD->var_offset + MD->vo_time_remaining, "--:--");


	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	free(MD);
	if(MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	int seconds;
	char *state = NULL;
	float part;
	char str[32];

	DBG_MSG(1, "Reading battery ...");

	MD->mod_timeout_value = MD->refresh * 1e3;

	if (read_battery(MD->battery, &seconds, &part, &state)) {
		var_set_as_float(MD->var_offset + MD->vo_float, part);

		if (state[0] == 0) {
			var_set_as_string(MD->var_offset + MD->vo_percent, "");
			var_set_as_int(MD->var_offset + 0, 0);
		} else {
			snprintf(str, 32, "%0.1f %%", part * 100.0);
			var_set_as_string(MD->var_offset + MD->vo_percent, str);
			var_set_as_int(MD->var_offset + 0, 1);
		}
		var_set_as_string(MD->var_offset + MD->vo_state_string, state);

		if (seconds >= 0) {
			if (MD->show_hours && seconds >= 3600) {
				snprintf(str, 32, "%d:%02d:%02d", seconds / 3600, (seconds / 60) % 60, seconds % 60);
			} else {
				snprintf(str, 32, "%02d:%02d", seconds / 60, seconds % 60);
			}
			var_set_as_string(MD->var_offset + MD->vo_time_remaining, str);
		} else {
			var_set_as_string(MD->var_offset + MD->vo_time_remaining, "--:--");
		}
	}

	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	/* unused */
	return 0;
}

