/*
 * main.c - icebuteo module "Filepool"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

#include "filepool.h"

/* debug */
#define DEBUG 0
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and one integer parameter.
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module real name */
const char *mod_name = "Filepool";

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */
	char *buffer;
	int buffer_size;

	/* configuration */
	char *filename;
	int execute_in_shell;
	int max_lines;
	int refresh;
	int size;
	float fraction_of;
	int trim;
	int vo_content, vo_fraction_of;
} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))


/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}

/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	int i = 1;
	P_TRANS cfg_tr[] = {
		/* const. name,			id, var, type */
		{ "execute-in-shell",		1, NULL, P_INTEGER },
		{ "filename",			2, NULL, P_STRING },
		{ "fraction-of",		3, NULL, P_FLOAT },
		{ "max-lines",			4, NULL, P_INTEGER },
		{ "refresh",			4, NULL, P_INTEGER },
		{ "size",			5, NULL, P_INTEGER },
		{ "trim",			6, NULL, P_INTEGER },
		{ "var-content",		7, NULL, P_INTEGER },
		{ "var-fraction-of",		8, NULL, P_INTEGER },
		{ NULL, 0, NULL, 0 }
	};
	const int cfg_tr_len = sizeof(cfg_tr) / sizeof(cfg_tr[0]) - 1;
	static P_FORMAT cfg_format[] = {
		P_CONSTANT, P_EQUAL, P_STRING | P_INTEGER | P_FLOAT, 0
	};

	DBG_MSG(1, "initialize...");
	DBG_MSG(3, "cfg_tr_len = %d", cfg_tr_len);

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));
	memset((void *)*mod_data, 0, sizeof(MOD_DATA));

	/* save offset */
	MD->var_offset = var_offset;

	/* load configuration */
	i = 0;
	cfg_tr[i++].var = &(MD->execute_in_shell);
	cfg_tr[i++].var = &(MD->filename);
	cfg_tr[i++].var = &(MD->fraction_of);
	cfg_tr[i++].var = &(MD->max_lines);
	cfg_tr[i++].var = &(MD->refresh);
	cfg_tr[i++].var = &(MD->size);
	cfg_tr[i++].var = &(MD->trim);
	cfg_tr[i++].var = &(MD->vo_content);
	cfg_tr[i++].var = &(MD->vo_fraction_of);
	if (!parse_file(arg, NULL, cfg_format, -1, 0, 2, cfg_tr, cfg_tr_len)) {
		return -1;
	}

	/* register timeout - first after 0.15 seconds */
	MD->mod_timeout_value = 150e3;

	/* init vars */
	var_set_as_int(MD->var_offset + 0, 1);
	var_set_as_string(MD->var_offset + MD->vo_content, "(nothing yet)");
	
	DBG_MSG(3, "File: \"%s\"", MD->filename);
	DBG_MSG(3, "Size: %d", MD->size);
	DBG_MSG(3, "Variable: %d", MD->var_offset + MD->vo_content);

	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	if(MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	if(MD->buffer)
		free(MD->buffer);
	free(MD);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	char *buf, *p_buf, *p_line, *p_eol, *p_trim, *p_next_line;
	float fract;
	int line = 0;
	
	DBG_MSG(1, "Reading file %s ...", MD->filename);

	MD->mod_timeout_value = MD->refresh * 1e3;

	if ((buf = read_file(MD->filename, MD->size, MD->execute_in_shell, &(MD->buffer), &(MD->buffer_size)))) {
		DBG_MSG(3, "File content: \"%s\"", buf);
		p_next_line = buf;
		while (*p_next_line && line < MD->max_lines) {
			/* set to begin and end of line */
			p_line = p_next_line;
			p_eol = strchr(p_line, '\n');

			/* mark eol */
			if (p_eol) {
				*p_eol = 0;
				p_next_line = p_eol + 1;
			} else {
				p_eol = p_buf + strlen(p_buf);
				p_next_line = NULL;
			}
			
			DBG_MSG(4, "RAW line: \"%s\"", p_line);
			
			/* trim whitespace and set p_buf to line real begin */
			if (MD->trim) {
				/* trim from end */
				for(p_trim = p_eol - 1; p_trim > p_line && isspace(*p_trim); p_trim--)
					;
				*(p_trim+1) = 0;

				/* trim from begin */
				for(p_buf = p_line; *p_buf && isspace(*p_buf); p_buf++)
					;
				DBG_MSG(4, "Trimmed line: \"%s\"", p_buf);
			} else {
				p_buf = p_line;
			}

			/* store value */
			DBG_MSG(3, "Line: \"%s\"", p_buf);
			var_set_as_string(MD->var_offset + MD->vo_content + line, p_buf);

			/* calculate fraction from 1st line */
			if (line == 0 && MD->fraction_of) {
				fract = atof(p_buf) / MD->fraction_of;
				var_set_as_float(MD->var_offset + MD->vo_fraction_of, fract);
				DBG_MSG(3, "Fraction of %f: %f", MD->fraction_of, fract);
			}
			line++;

			/* set visibility var */
			var_set_as_int(MD->var_offset, 1);
		}
	} else {
		DBG_MSG(3, "No content.");
		var_set_as_int(MD->var_offset, 0);
	}

	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	/* unused */
	return 0;
}

