/*
 * main.c - icebuteo module "WiFi"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

#include "wm_arch.h"

/* debug */
#define DEBUG 0
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and one integer parameter.
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module real name */
const char *mod_name = "WiFi";

#define FIFO_BUF_MAX	128

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */

	/* configuration */
	float update_seconds;
	char *interface;
	int var_iface;
	int var_essid;
	int var_ap;
	int var_bitrate;
	int var_linkquality_str;
	int var_linkquality_bar;
} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))


/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}

/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	int i = 1;
	P_TRANS cfg_tr[] = {
		/* const. name,			id, var, type */
		{ "interface",			1, NULL, P_STRING },
		{ "update-seconds",		2, NULL, P_FLOAT },
		{ "var-ap",			3, NULL, P_INTEGER },
		{ "var-bitrate",		4, NULL, P_INTEGER },
		{ "var-essid",			5, NULL, P_INTEGER },
		{ "var-iface",			6, NULL, P_INTEGER },
		{ "var-linkquality-bar",	7, NULL, P_INTEGER },
		{ "var-linkquality-str",	8, NULL, P_INTEGER },
		{ NULL, 0, NULL, 0 }
	};
	const int cfg_tr_len = sizeof(cfg_tr) / sizeof(cfg_tr[0]) - 1;
	static P_FORMAT cfg_format[] = {
		P_CONSTANT, P_EQUAL, P_STRING | P_INTEGER | P_FLOAT, 0
	};

	DBG_MSG(1, "initialize...");

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));
	memset((void *)*mod_data, 0, sizeof(MOD_DATA));
	MD->mod_filedescriptor_value = NULL;

	/* save offset */
	MD->var_offset = var_offset;

	/* load configuration */
	i = 0;
	cfg_tr[i++].var = &(MD->interface);
	cfg_tr[i++].var = &(MD->update_seconds);
	cfg_tr[i++].var = &(MD->var_ap);
	cfg_tr[i++].var = &(MD->var_bitrate);
	cfg_tr[i++].var = &(MD->var_essid);
	cfg_tr[i++].var = &(MD->var_iface);
	cfg_tr[i++].var = &(MD->var_linkquality_bar);
	cfg_tr[i++].var = &(MD->var_linkquality_str);
	if (!parse_file(arg, NULL, cfg_format, -1, 0, 2, cfg_tr, cfg_tr_len)) {
		return -1;
	}

	/* register timeout - first after 0.1 seconds */
	MD->mod_timeout_value = 100e3;
	MD->mod_timeout_always = 0;

	/* init vars */
	var_set_as_int(MD->var_offset + 0, 1);
	var_set_as_string(MD->var_offset + MD->var_iface, MD->interface);
	var_set_as_string(MD->var_offset + MD->var_ap, "00:00:00:00:00:00");
	var_set_as_string(MD->var_offset + MD->var_essid, "(none)");
	var_set_as_string(MD->var_offset + MD->var_bitrate, "0 Mbps");
	var_set_as_float(MD->var_offset + MD->var_linkquality_bar, 0.0);
	var_set_as_string(MD->var_offset + MD->var_linkquality_str, "--");

	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	free(MD);
	if(MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	int sock, q_cur = 0, q_max = 0, bitrate, visible = 0;
	char buf[128];
	float lq;
	
	MD->mod_timeout_value = MD->update_seconds * 1e6;

	sock = wm_open_socket(MD->interface);
	
	if (sock) {
		/* set visibility */
		visible = wm_check(sock, MD->interface);
		var_set_as_int(MD->var_offset, visible);

		if (visible) {
			/* essid */
			wm_get_essid(sock, MD->interface, buf, sizeof(buf));
			var_set_as_string(MD->var_offset + MD->var_essid, buf);
			DBG_MSG(2, "essid: \"%s\"", buf);
			
			/* link quality */
			switch (wm_get_quality(sock, MD->interface, &lq, &q_cur, &q_max)) {
				case -1:
					strcpy(buf, "(err)");
					break;
				case 0:
					strcpy(buf, "--");
					break;
				default:
					snprintf(buf, sizeof(buf) - 1, "%d / %d", q_cur, q_max);
					buf[sizeof(buf) - 1] = 0;
					break;
			}
			var_set_as_float(MD->var_offset + MD->var_linkquality_bar, lq);
			var_set_as_string(MD->var_offset + MD->var_linkquality_str, buf);
			DBG_MSG(2, "Link Quality: %f, \"%s\"", lq, buf);

			/* bitrate */
			bitrate = wm_get_bitrate(sock, MD->interface);
			snprintf(buf, sizeof(buf) - 1, "%0.1f Mbps", bitrate / 1000000.);
			var_set_as_string(MD->var_offset + MD->var_bitrate, buf);
			DBG_MSG(2, "Bitrate: %s", buf);

			/* AP */
			wm_get_ap(sock, MD->interface, buf, sizeof(buf));
			var_set_as_string(MD->var_offset + MD->var_ap, buf);
			DBG_MSG(2, "AP: \"%s\"", buf);
		}
		
		/* cleanup */
		wm_close_socket(sock);
	}

	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	DBG_MSG(1, "File descriptor callback...");
	DBG_MSG(2, "File descriptor callback... Done.");
	return 0;
}

