/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 * This code is from kwirelessmonitor.
 *
 * Modified by Josef Kufner <jk@myserver.cz>
 * 
 */

#if defined(__linux__)
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <linux/wireless.h>	/* wireless extension */

#include "wm_arch.h"

int wm_open_socket(const char *ifname)
{
	if (strlen(ifname) < 1) {
		return -1;
	}
	return socket(PF_INET, SOCK_DGRAM, 0);
}


void wm_close_socket(int sock)
{
	if (sock >= 0) {
		close(sock);
	}
}


int wm_check(int sock, const char *ifname)
{
	struct iwreq wreq;
	
	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);

	return (ioctl(sock, SIOCGIWNAME, &wreq) == 0);
}

void wm_get_essid(int sock, const char *ifname, char *buf, int len)
{
	struct iwreq wreq;
	
	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);
	wreq.u.essid.pointer = (caddr_t) buf;
	wreq.u.essid.length = len;
	wreq.u.essid.flags = 0;
	
	if (ioctl(sock, SIOCGIWESSID, &wreq) < 0) {
		strncpy(buf, "unknown", len - 1);
		buf[len] = 0;
	} else {
		buf[wreq.u.essid.length] = 0;
	}
}


int wm_get_bitrate(int sock, const char *ifname)
{
	struct iwreq wreq;
	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);
	if (ioctl(sock, SIOCGIWRATE, &wreq) < 0) {
		return -1;
	}
	return wreq.u.bitrate.value;
}


void wm_get_ap(int sock, const char *ifname, char *buf, int len)
{
	struct iwreq wreq;
	int i;
	
	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);

	if (ioctl(sock, SIOCGIWAP, &wreq) < 0) {
		strncpy(buf, "unknown", len - 1);
		buf[len] = 0;
	} else {
		for (i = 0; i < sizeof(wreq.u.ap_addr.sa_data) && i < 6 && i * 4 < len; i++) {
			snprintf(buf + i * 3 - (buf[0] != 0), 4, "%s%02X", buf[0] ? ":":"", (int)(wreq.u.ap_addr.sa_data[i]));
		}
	}
}


int wm_get_max_quality(int sock, const char *ifname)
{
	char buffer[sizeof(struct iw_range) * 2];
	struct iwreq wreq;

	memset(buffer, 0, sizeof(buffer));
	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);

	wreq.u.data.pointer = (caddr_t) buffer;
	wreq.u.data.length = sizeof(buffer);
	wreq.u.data.flags = 0;
	if (ioctl(sock, SIOCGIWRANGE, &wreq) < 0) {
		return 1;
	}

	struct iw_range *wrange = (struct iw_range *) buffer;
	return wrange->max_qual.qual;
}


int wm_get_quality(int sock, const char *ifname, float *link_quality, int *cur_quality, int *max_quality)
{
	/* Get interface stats */
	/* (Assume WIRELESS_EXT > 11) */
#if WIRELESS_EXT < 17
#	error "WIRELESS_EXT >= 17 required"
#endif
	struct iwreq wreq;
	struct iw_statistics wstats;

	strncpy(wreq.ifr_name, ifname, IFNAMSIZ);
	wreq.u.data.pointer = (caddr_t) & wstats;
	wreq.u.data.length = sizeof(struct iw_statistics);
	wreq.u.data.flags = 1;
	if (ioctl(sock, SIOCGIWSTATS, &wreq) < 0) {
		perror("wm_get_quality");
		return -1;
	}

	int max_qual = wm_get_max_quality(sock, ifname);
	int quality = wstats.qual.qual;
	int flag = wstats.qual.updated;

	if (max_quality)
		*max_quality = (int) max_qual;

	if (flag & IW_QUAL_QUAL_UPDATED) {
		if (link_quality)
			*link_quality = log((double) quality) / log((double) max_qual);
		if (cur_quality)
			*cur_quality = (int) quality;

		return 1;
	} else {
		if (link_quality)
			*link_quality = 0;
		if (cur_quality)
			*cur_quality = (int) 0;

		return 0;
	}
}


void wm_get_transfer_bytes(const char *ifname, unsigned long long *rx_bytes, unsigned long long *tx_bytes)
{
	FILE *f = fopen("/proc/net/dev", "r");
	if (f == NULL) {
		*rx_bytes = 0;
		*tx_bytes = 0;
		return;
	}

	char buf[512];
	fgets(buf, sizeof(buf), f);
	fgets(buf, sizeof(buf), f);
	char *stats = NULL;
	while (fgets(buf, sizeof(buf), f)) {
		char *begin = buf;
		while (isspace(*begin)) {
			begin += 1;
		}
		char *end = strrchr(buf, ':');
		if (end) {
			if (strncmp(begin, ifname, strlen(ifname))) {
				continue;
			}
			stats = end + 1;
			while (isspace(*stats)) {
				stats += 1;
			}
			break;
		}
	}
	fclose(f);

	if (stats) {
		sscanf(stats, "%llu %*u %*u %*u %*u %*u %*u %*u %llu ", rx_bytes, tx_bytes);
	} else {
		*rx_bytes = 0;
		*tx_bytes = 0;
	}
}

#endif				/* defined(__linux__) */
